/**
 * Node.js file as the server to accept Drupal data.
 * File upload: use formidable example (https://github.com/felixge/node-formidable)
 *
 * This will only work with MySQL database.
 * Database needs to have the {async_command} table and reccloud_client table, defined as: nid, uid, apikey, ready.
 *
 */

// set required libraries.
var http = require('http'),
    util = require('util'),
    path = require('path'),
    formidable = require('formidable'),
    paperboy = require('paperboy'),
    mysql = require('mysql'),
    iniparser = require('iniparser');


// set default parameters
var config = iniparser.parseSync('config.properties');
var default_port = config['node_port'] != undefined ? config['node_port'] : 2012;
var data_folder = ('node_data_folder' in config) ? config['node_data_folder'] : (process.cwd() + '/' + 'data');

// init database
var database_host = ('host_name' in config) ? config['host_name'] : 'localhost';
var database_port = ('host_port' in config) ? config['host_port'] : 3306;
var db = mysql.createClient({
  user: config['username'],
  password: config['password'],
  host: database_host,
  port: database_port,
  database: config['database_name']
});

// helper functions
function respond_message(res, http_code, message) {
  res.writeHead(http_code, {'content-type': 'text/plain'});
  res.end(message);
}

function respond_json(res, http_code, data) {
  res.writeHead(http_code, {'content-type': 'text/json'});
  res.end(JSON.stringify(data));
}

function _test_node(res) {
  // test db query
  db.query('SELECT ready FROM reccloud_client WHERE apikey=?', [1234567890123456], function (err, results, fields) {
    if (err) {
      throw err;
    }
    console.log(results);
    //console.log(fields);
    db.end();
  });

  res.writeHead(200, {'content-type': 'text/html'});
  res.end(
    '<form action="/upload" enctype="multipart/form-data" method="post">' +
      '<input type="text" name="title"><br>' +
      '<input type="text" name="description"><br>' +
      '<input type="file" name="upload" multiple="multiple"><br>' +
      '<input type="submit" value="Upload">' +
      '</form>'
  );
}


function _handle_upload(res, req) {
  var files = [], fields = [];
  var form = new formidable.IncomingForm();
  var recommender_params;

  form.keepExtensions = true;
  form
    .on('field', function(field, value) {
      //console.log(field, value);
      if (field == 'recommender_params') {
        recommender_params = value;
      }
      fields.push([field, value]);
    })
    .on('file', function(field, file) {
      //console.log(field, file);
      console.log('-> upload request: ' + file.name);
      files.push([field, file]);
    })
    .on('fileBegin', function(name, file) {
      // FIXME: need to validate file name in order to determine availability, also prevent malicious update to break the system.
      var filename = path.basename(file.name, path.extname(file.name));   // TODO: might not need the extname()
      var parts = filename.split('-', 2);
      var apikey = parts[0], command_id = parts[1];

      // name is the form name.
      file.path = data_folder + '/' + path.basename(file.name);

      db.query('SELECT nid, uid FROM reccloud_client WHERE apikey=?', [apikey])
        .on('row', function(row) {
          var nid = row['nid'];
          var uid = row['uid'];
          // only insert the command when the API key is valid.
           db.query('INSERT INTO async_command(app, command, description, eid, uid, created, string1, string2, string3, id1) VALUES (?, ?, ?, ?, ?, UNIX_TIMESTAMP(), ?, ?, ?, ?)',
              ['reccloud', 'RunRemoteRecommender', 'Request to run remote recommender.', nid, uid, apikey, file.path, recommender_params, command_id]);
        });
    })
    .on('end', function() {
      // console.log('-> upload done');
      res.writeHead(200, {'content-type': 'text/plain'});
      res.write('received fields:\n\n ' + util.inspect(fields));
      res.write('\n\n');
      res.end('received files:\n\n ' + util.inspect(files));
    })
    .on('error', function(err) {
      console.log('-> upload error: ' + err);
      respond_message(res, 400, 'Unknown bad request: ' + err);
    });

  form.parse(req);
}


function _handle_check_ready(res, req, apikey) {
  var ready = 0;
  db.query('SELECT ready FROM reccloud_client WHERE apikey=?', [apikey])
    .on('row', function(row) {
      //console.log(row);
      ready = row['ready'];
    })
    .on('end', function() {
      if (ready == 1) {
        respond_json(res, 200, {'message': 'Service ready.'});
      } else {
        respond_json(res, 401, {'message': 'Service denied. Possible cause: API key error or your quota is reached.'});
      }
    });
}

function _handle_check_status(res, req, apikey, id) {
  var found = false, status, message;
  db.query('SELECT status, message FROM async_command WHERE string1=? AND id1=? AND app="reccloud" AND command="RunRemoteRecommender" ORDER BY created DESC LIMIT 1', [apikey, id])
    .on('row', function(row) {
      //console.log(row);
      found = true;
      status = row['status'];
      message = row['message'];
    })
    .on('end', function() {
      if (found) {
        var data = {};
        if (status != undefined) {
          data['status'] = status;
          data['message'] = message;
        } else if (message == undefined) {
          data['message'] = 'Command execution is queued, waiting to be handled';
        }
        respond_json(res, 200, data);
      } else {
        respond_json(res, 404, {'message': 'The command execution request is not found in remote server.'});
      }
    });
}



function _handle_check(res, req) {
//console.log(util.inspect(req));
  var form = new formidable.IncomingForm();
  var fields = {};

  form
    .on('field', function(field, value) {
      //console.log(field, value);
      fields[field] = value;
    })
    .on('end', function() {
      // console.log('-> upload done');
      //console.log(util.inspect(fields));
      //console.log(util.inspect(files));
      if ('apikey' in fields && 'what' in fields) {
        console.log('-> check request: ' + fields['what'] + ' - ' + fields['apikey']);
        if (fields['what'] == 'ready') {
          _handle_check_ready(res, req, fields['apikey']);
        }
        else if (fields['what'] == 'status') {
          _handle_check_status(res, req, fields['apikey'], fields['id']);
        }
      }
      else {
        console.log('-> check error: missing apikey or the "what" to check');
        respond_message(res, 400, 'missing apikey or the "what" to check.');
      }
    })
    .on('error', function (err) {
      console.log('-> check error: ' + err);
      respond_message(res, 400, 'Unknown bad request: ' + err);
    });

  form.parse(req);
}


function _handle_download(req, res) {
  req.url = req.url.substring(9); // strip '/download'
  console.log('-> download request: ' + req.url);
  paperboy.deliver(data_folder, req, res);
}


function _handle_404(req, res) {
  console.log('-> error request: ' + req.url);
  res.writeHead(404, {'content-type': 'text/plain'});
  res.end('Error request.');
}

var server = http.createServer(function(req, res) {
  if (req.url == '/') {
    // TODO: send default help.html
  } else if (req.url == '/test') {
    _test_node(res);
  } else if (req.url == '/check' && req.method == 'POST') {
    _handle_check(res, req);
  } else if (req.url == '/upload' && req.method == 'POST') {
    _handle_upload(res, req);
  } else if (req.url.indexOf('/download/') == 0) {
    _handle_download(req, res);
  } else {
    _handle_404(req, res);
  }
});


process.on('exit', function() {
  db.end();
});

server.listen(default_port);
console.log('listening on http://localhost:'+default_port+'. ');